## Illustrate linear regression
## This script stems from the lab accompanying Chapter 3 of Introduction to Statistical Learning

# collection of demo data sets and functions (we will use Boston dataset)
library("MASS") 
# tutorial data sets 
library("ISLR") 
library(tidyverse)

## use Boston data set available in MASS to predict median house value in neighborhoods around Boston
help(Boston) # learn the structure of the dataset, 506 areas described by 14 variables, medv will serve as the target variable

# Plot the relationship between lstat (the percent of low socioeconomic households) 
# and medv (median value of homes in $1000.)
Boston %>% 
  ggplot(aes(x=lstat, y=medv)) +
  geom_point()

## Simple linear regression
lm_fit <- lm(medv ~ lstat, data = Boston) 

# visualize the relationship between the independent and the dependent variable
Boston %>% 
  ggplot(aes(x=lstat, y=medv)) +
  geom_point() +
  # visualize the fit
  geom_abline(slope = lm_fit$coefficients[_____],
              intercept = lm_fit$coefficients[_____],
              size=1, col='red') +
  ggtitle("Relationship between lower status population and house prices")

summary(lm_fit) # show extensive metrics about the model

# ----------------------------------------- END OF THE 3rd LAB ----------------------------------------------

names(lm_fit)   # get all component names of the lm model
coef(lm_fit)    # see the coefficients once more, same as lm_fit$coefficients
confint(lm_fit) # confidence intervals around these estimates

# predict for three new neighborhoods including confident intervals
predict(lm_fit, 
        data.frame(lstat=c(5, 10, 15)), 
        interval = "confidence") 

par(mfrow=c(2, 2))
plot(lm_fit) # see the diagnostic plots

plot(hatvalues(lm_fit)) # Leverage statistics can be computed for our predictors with hatvalues()
which.max(hatvalues(lm_fit)) # find the observation with the largest leverage

## Summary
# There is an obvious relationship between the independent variable (household status) and the dependent variable (median house value).
# Both the F-stat for the whole model and t-stat for the predictor are significant, the predictor helps to explain more than 50% of the variance in the target variable.
# The median house price is around $34500, on average it decreases by $950 with one percent increase of households with low socioeconomical status. 
# For more than 30% it virtually becomes $0 (the model does not extrapolate well for these unobserved values).
# Both simple plot and residual plot suggest that the relationship is rather non-linear.
# Consequences: prediction could be improved with additional non-linear terms, assumptions are violated, estimations do not have to be perfect (namely p-values, confidence intervals).

nlm_fit <- lm(medv ~ poly(lstat,2), data = Boston) # just see the improvement with the most simple polynomial regression
summary(nlm_fit) # explains around 10% variance more
plot(nlm_fit) # residual plot looks much better (still not normal error distribution)
anova(lm_fit, nlm_fit) # the second model is significantly better, sum of sq is positive, p-value small
# We will look at ANOVA in next labs, but the concept is not new - it's an F-test that compares 2 models via their R2 scores

## Multiple linear regression
# Use multiple/all/relevant predictors at once, in a single model the available independent variables
lm_fit_all <- lm(medv ~ ., data=Boston)

summary(lm_fit_all) # model improves again, only some of the variables seem to be unimportant, we may want to exclude them

lm_fit_exclude <- lm(medv ~ . -age -indus, data=Boston) # truly exclude age and indus
summary(lm_fit_exclude) # the simpler model seems to maintain the performance of the previous model
anova(lm_fit_all,lm_fit_exclude) # no difference, the simpler model pref

## Finally, try some feature selection methods
step <- stepAIC(lm_fit_all, direction="both") # stepwise regression, based on AIC criterion, taken from MASS library
step$anova # display results, actually does the same as we did manually, removes age and indus

# Fit LASSO to remove unimportant predictors
library("glmnet") # lasso

lambda_grid <- 10^ seq(10, -3, length =200)
BostonX <- Boston %>% dplyr::select(-medv) %>% as.matrix()

lasso.model <- glmnet(x=BostonX,
                      y=Boston$medv,
                      alpha=1, 
                      lambda=lambda_grid, 
                      standardize=TRUE)
plot(lasso.model, "lambda", label = TRUE) # See regularization in action

# Lambda controls RMSE-shrinkage trade-off. Determine its 'optimal' value
lasso.cv.out <- cv.glmnet(x=BostonX,
                          y=Boston$medv,
                          alpha=1)
# Plot lambda vs. RMSE
plot(lasso.cv.out) 
lasso.lambda <- lasso.cv.out$lambda.min

# Obtain coefficients corresponding to given lambda
lasso.coefficients <- predict(lasso.model, type="coefficients", s=lasso.lambda)

# Display the coefficients and selected variables
print("LASSO coefficients:")
print(lasso.coefficients) # the absolute coefficient values influenced by the scale of the individual variables, nox has a small scale
print(as.matrix(lasso.coefficients)[seq(2,length(Boston)-1),] != 0) 

## Questions
# 1. What procedure is used to choose lambda?
# 2. Compare lasso.cv.out$lambda.min and lasso.cv.out$lambda.1se - find them in the plot(lasso.cv.out). Which one do you prefer and why?
# 3. How lasso differs from ridge and how would you perform ridge regression? (hint: look at what parameters were used in the glmnet function)

## Summary
# To avoid overfitting and increase simplicity and understandability of the model, only relevant features should be used.
# We have shown three different approaches to feature selection: p-values, stepwise regression and lasso. All of them came to the same conclusion: two features should be removed from the model.
